
import ecdsa

# Note that this is python ecdsa library-centric atm, but can easily be extended to support other
#      keys from other libraries

from c3.constants import KT_ECDSA_PRIME256V1, KT_ECDSA_SECP256K1
from c3.errors import SignError

# NIST P-256 aka secp256r1 aka prime256v1
# SECP256k1 aka the bitcoin one

# --- C3 curve constants to python ecdsa library curve objects ---
C3_KT_TO_PY_ECDSA_CURVE = {
    KT_ECDSA_PRIME256V1 : ecdsa.NIST256p,
    KT_ECDSA_SECP256K1  : ecdsa.SECP256k1,
}

def c3keytype_to_ecdsa_curve(keytype, task_msg):
    if keytype not in C3_KT_TO_PY_ECDSA_CURVE:
        raise NotImplementedError("Error %s - unknown keytype" % (task_msg,))
    return C3_KT_TO_PY_ECDSA_CURVE[keytype]


def generate(keytype):
    curve = c3keytype_to_ecdsa_curve(keytype, "generating keypair")
    priv = ecdsa.SigningKey.generate(curve=curve)
    pub = priv.get_verifying_key()
    return priv.to_string(), pub.to_string()

def sign_make_sig(keytype, priv_bytes, payload_bytes):
    curve = c3keytype_to_ecdsa_curve(keytype, "signing payload")
    SK = ecdsa.SigningKey.from_string(priv_bytes, curve)
    sig_bytes = SK.sign(payload_bytes)
    return sig_bytes

def verify(cert, payload_bytes, signature_bytes):
    curve = c3keytype_to_ecdsa_curve(cert.key_type, "verifying payload")
    VK = ecdsa.VerifyingKey.from_string(cert.public_key, curve)
    return VK.verify(signature_bytes, payload_bytes)  # returns True or raises exception

def check_privpub_match(cert, priv_key_bytes):
    curve = c3keytype_to_ecdsa_curve(cert.key_type, "checking keypair match")
    priv = ecdsa.SigningKey.from_string(priv_key_bytes, curve)
    pub = priv.get_verifying_key()
    if pub.to_string() != cert.public_key:
        raise SignError("Private key and public key do not match")
    return True





# C3 Signing and verifying actions, and a little in-ram trust store for verify().

import datetime, functools, operator

from c3.constants import *
from c3.errors import *
from c3 import certentry, keypairs
from c3 import structure, textfiles
from c3 import pass_protect, parsedate
from c3.structure import AttrDict

# Policy: anything with "block" in the name is bytes.
#         File, Txt, Block

# Policy: public binary structure - sticking with "chain[0].data_part can be payload or cert"
#         rather than "seperate payload & sig top level structure" because it makes verify simpler
#         & its too much change at this point.
#         (Also makes verify a lot simpler to implement in *other* languages quickly)

# Policy: verify() only reads from self.trusted_ces, it doesnt write anything into there.
#         Caller/User must call add_trusted_certs() to add theirs.

# Note: the priv key encrypt/decrypt functions are here too, just to keep things to 1 class for now.

class SignVerify(object):
    def __init__(self):
        self.trusted_ces = {}   # by name. Used by verify().

        # This uses libsodium so we load set it up at startup so the DLL can load at startup
        self.pass_protect = pass_protect.PassProtect()


    # ============ Load  ==================================================================

    # Note: vis_map is ONLY for text-stuff, the binary stuff doesn't actually care _ever_ about the
    #       user's schema. The USER does, after the user get_payloads.

    # I think the file loaders can now be really basic.
    # Because all the smarts is in the text processor right here below.
    # Just open the files and read them and dump them into a single text variable. The filenames
    # dont actually matter to us, what drives public/private is the "PRIVATE" in the header line.
    # (that's also the only text part that controls anything, everything else is in the binary blocks.)

    # Policy: not supporting Visible Fields for the private block atm.
    #         The private block doesn't have a subject name anyway, we're relying on keypair crosscheck

    def load(self, filename="", text="", block=b"", vis_map=None):
        highlander_check(filename, text, block)  # there can be only one of these 3
        ce = certentry.CertEntry(self)
        pub_vf_lines = ""
        payload_dict = {}

        # Note: this if-flow works because text_file, text, and block are mutually exclusive
        # --- LOAD from aguments ---
        if filename:
            text = textfiles.load_files(filename)

        if text:  # Text is EITHER, public text, private text, or both texts concatenated.
            ce.pub_text, ce.epriv_text = textfiles.split_text_pub_priv(text)
            if ce.pub_text:
                ce.pub_block, pub_vf_lines = textfiles.text_to_binary_block(ce.pub_text)
            if ce.epriv_text:
                ce.epriv_block, _ = textfiles.text_to_binary_block(ce.epriv_text)
                # Note: ignoring vf_lines for private text atm.

        if block:
            ce.pub_block, ce.epriv_block = structure.split_binary_pub_priv(block)

        # --- Unpack binary blocks ---
        if not ce.pub_block:            # only bare-payload "CE"s dont have a public part, and they
            raise ValueError("Load: public block is missing")  # dont come through here, so.

        if ce.epriv_block:
            ce.priv_d = structure.load_priv_block(ce.epriv_block)
            if ce.priv_d.priv_type == PRIVTYPE_BARE:    # load automatically if the supplier
                ce.priv_key_bytes = ce.priv_d.priv_data  # didnt care about securing it.
            # if it IS encrypted, do nothing. Caller must call decrypt(), otherwise sign() will fail.

        ce.pub_type, thingy = structure.load_pub_block(ce.pub_block)

        if ce.pub_type == PUB_CSR:      # CSRs are just a cert
            ce.cert = thingy                # we're mixing cert-level stuff with CE-level stuff
            ce.name = ce.cert.subject_name      # noqa a bit here, so there are double-ups.
            if pub_vf_lines:        # tamper check public Visible Fields if any
                textfiles.crosscheck_visible_fields(pub_vf_lines, ce.default_vismap, ce.cert)

        if ce.pub_type == PUB_CERTCHAIN:
            ce.chain = thingy
            ce.name = ce.chain[0].cert.subject_name
            ce.cert = ce.chain[0].cert
            if pub_vf_lines:        # tamper check public Visible Fields if any
                textfiles.crosscheck_visible_fields(pub_vf_lines, ce.default_vismap, ce.cert)

        if ce.pub_type == PUB_PAYLOAD:          # note: no name, no cert
            ce.chain = thingy
            ce.payload = ce.chain[0].data_part
            # tamper check user Visible Fields if any. User-supplied schema is required.
            if pub_vf_lines:
                if not vis_map or "schema" not in vis_map or not vis_map["schema"]:
                    helpm = ". (please supply vis_map= to load() function)"
                    raise StructureError("Payload has visible fields but schema unknown" + helpm)
                payload_dict = AttrDict(b3.schema_unpack(vis_map["schema"], ce.payload))
                textfiles.crosscheck_visible_fields(pub_vf_lines, vis_map, payload_dict)

        return ce

    def load_trusted_cert(self, *args, **kw):
        ce = self.load(*args, **kw)
        force = "force" in kw and kw["force"] is True
        if not force:
            try:
                self.verify(ce)
            except UntrustedChainError:  # ignore this one failure mode because we havent installed
                pass                     # this/these certs yet
        # add to registry
        self.trusted_ces[ce.cert.cert_id] = ce            # Note: by ID
        return

    # ============ Makers ===============

    # Note: CSRs are exportable, they have their own binary format which is just the cert.
    #       so not a chain like all the others.
    #       bare_payloads are NOT exportable, they are intended to be signed immediately.
    #       we want CSR loads to be different to "just sign a payload" because we want to have the
    #       option of e.g. adjusting the wanted expiry date, etc.

    def make_csr(self, name, expiry, cert_type=None):
        expiry = parsedate.ParseBasicDate(expiry)
        ce = certentry.CertEntry(self)
        ce.pub_type = PUB_CSR
        ce.name = name

        #cert_id = name.encode("ascii")  # if we want deterministic cert_ids e.g. for testing
        cert_id = structure.gen_ulid()
        today = datetime.date.today()
        key_priv, key_pub = keypairs.generate(keytype=KT_ECDSA_SECP256K1)

        ce.cert = AttrDict(public_key=key_pub, subject_name=name, cert_id=cert_id, issued_date=today,
                           key_type=KT_ECDSA_SECP256K1, expiry_date=expiry, cert_type=cert_type)

        ce.priv_key_bytes = key_priv
        # Note: we don't set ce.epriv_block here, user must call encrypt() or nopassword() to make
        #       that happen, and we wont save without one of them happening.

        # All the exportable (as_binary etc) pub_blocks are serializations of ce.chain, except for
        # CSRs, which are serializations of ce.cert.
        cert_block = b3.schema_pack(CERT_SCHEMA, ce.cert)
        ce.pub_block = b3.encode_item_joined(PUB_CSR, b3.DICT, cert_block)
        return ce

    def make_payload(self, payload_bytes):
        ce = certentry.CertEntry(self)
        ce.pub_type = BARE_PAYLOAD
        ce.payload = payload_bytes
        return ce


    # =========== Sign ==================

    # so self-sign is make_csr, sign, encrypt_priv, save
    # non-self-sign is load, load_signer, sign, encrypt_priv if not already, save

    def sign(self, ce, signer, link_by_name=False):        # link_name
        bytes_to_sign = b""
        payload = b""
        self_signing = (ce == signer)
        # Take cert, turn into bytes, get privkey from signing, sign bytes, make sig.
        # Then repatch to_sign's pub_block with signing_ce's pub_block.

        if datetime.date.today() > signer.cert.expiry_date:
            raise CertExpired("Signing cert has expired")

        # First param needs to always be a ce if we are in-place transforming it.
        # Either we are signing payload bytes, or we are signing cert bytes.
        if ce.cert:
            payload = b3.schema_pack(CERT_SCHEMA, ce.cert)
        if ce.payload:
            payload = ce.payload
        if not payload:
            raise SignError("No cert or payload found to sign")
        if not signer.priv_key_bytes:
            raise SignError("Please decrypt the signer's private key first")
        keypairs.check_privpub_match(signer.cert, signer.priv_key_bytes)
        self.ensure_not_expired(signer)
        # Note: we can't gate on existence of epriv_block first, because e.g. selfsigned CSRs dont have one.

        # perform sign with key, get signature bytes
        key_type = signer.cert.key_type
        sig_bytes = keypairs.sign_make_sig(key_type, signer.priv_key_bytes, payload)

        # build our chain with 'payload'&sig + signers chain
        signer_cert_id = signer.cert.cert_id if (link_by_name or self_signing) else b""
        datasig = self.make_datasig(payload, sig_bytes, signer_cert_id)

        ce.chain = [datasig]
        if not (link_by_name or self_signing):
            ce.chain += signer.chain

        if ce.pub_type == PUB_CSR:
            ce.pub_type = PUB_CERTCHAIN
        if ce.pub_type == BARE_PAYLOAD:
            ce.pub_type = PUB_PAYLOAD

        self.make_chain_pub_block(ce)     # serialize the chain

    def ensure_not_expired(self, signer_ce):
        expiry = signer_ce.cert["expiry_date"]
        if datetime.date.today() > expiry:
            raise CertExpired("Signing cert has expired")
        return True

    # --- Binary operations for sign() ---

    def make_datasig(self, payload_bytes, sig_bytes, signer_cert_id):
        sig_d = AttrDict(signature=sig_bytes, signing_cert_id=signer_cert_id)
        sig_part = b3.schema_pack(SIG_SCHEMA, sig_d)
        datasig = AttrDict(data_part=payload_bytes, sig_part=sig_part)
        return datasig

    def make_chain_pub_block(self, ce):
        # We need to pack the datasigs, then join those blocks, prepend the header, and that's pub_block
        chain_blocks = []
        for das in ce.chain:
            das_bytes = b3.schema_pack(DATASIG_SCHEMA, das)
            das_bytes_with_hdr = b3.encode_item_joined(HDR_DAS, b3.DICT, das_bytes)
            chain_blocks.append(das_bytes_with_hdr)
        ce.pub_block = b3.encode_item_joined(ce.pub_type, b3.LIST, b"".join(chain_blocks))



    # after sign() the pub_block is made, and if ce_encrypt is called, the priv block is made too?
    # The test can

    # ============ Verify ==========================================================================

    # In: Data-And-Sigs items list
    # Out: payload bytes or an Exception

    # Apart from actual signature fails, there are 3 other ways for this to fail:
    # 1) unnamed issuer cert and no next cert in line aka "fell off the end" (ShortChainError)
    # 2) Named cert not found - in the cert store / trust store / certs_by_name etc
    # 3) Last cert is self-signed and verifies OK but isn't in the trust store. (Untrusted Chain)

    # Note: CEs to verify MUST come from load() (ie not directly from make_csr+sign when testing)
    #       because load() fully unpacks the chain for verify to inspect.

    def verify(self, ce):
        chain = ce.chain
        ce.vcerts = []      # Computed trail, includes trusted_certs
        if not chain:
            raise StructureError("Cannot verify - no cert chain present")
        if "sig" not in chain[0]:
            raise StructureError("Cannot verify - CE must be load()ed first")

        certs_by_id = {das.cert.cert_id : das.cert for das in chain if "cert" in das}
        found_in_trusted = False         # whether we have established a link to the trusted_ces

        for i, das in enumerate(chain):
            if "cert" in das:
                ce.vcerts = [das.cert] + ce.vcerts

            # --- Find the 'next cert' ie the one which verifies our signature ---
            signing_cert_id = das.sig.signing_cert_id
            if not signing_cert_id:
                # --- no signing-id means "next cert in the chain" ---
                if i + 1 >= len(chain):      # have we fallen off the end?
                    raise ShortChainError(structure.ctnm(das)+"Next issuer cert is missing")
                next_cert = chain[i + 1].cert
            else:
                # --- got a name, look in trusted for it ---
                if signing_cert_id in self.trusted_ces:
                    tr_ce = self.trusted_ces[signing_cert_id]
                    next_cert = tr_ce.cert
                    ce.vcerts = tr_ce.vcerts + ce.vcerts
                    found_in_trusted = True
                # --- otherwise look in our own supplied chain ---
                elif signing_cert_id in certs_by_id:
                    next_cert = certs_by_id[signing_cert_id]
                else:
                    raise CertNotFoundError(structure.ctnm(das)+"signing cert not found (id %r)" % signing_cert_id)

            # --- Actually verify the signature ---
            try:
                keypairs.verify(next_cert, das.data_part, das.sig.signature)
            except Exception:       # wrap theirs with our own error class
                raise InvalidSignatureError(structure.ctnm(das)+"Signature failed to verify")

            # --- Now do next das in line ---

        # Chain verifies completed without problems. Make sure we got to a trust store cert.
        # If there is a payload, return it otherwise return True. (So bool-ness works.)
        # Even tho all errors are exceptions.
        if found_in_trusted:
            return True

        raise UntrustedChainError("Chain does not link to trusted certs")




# Policy: Arguments are mutually-exclusive,
#         not more and not less than one argument must have a value.
def highlander_check(*args):
    ibool_args = [int(bool(i)) for i in args]
    num_true = functools.reduce(operator.add, ibool_args, 0)
    if num_true == 0:
        raise ValueError("Please specify one mandatory argument (none or empty specified)")
    if num_true > 1:
        raise ValueError("Please specify only one mandatory argument (multiple were specified)")
    return True

# Policy:
# Using the "Store everything in a smart object" approach so that we can get the **APIs usable**
# Because the library user not losing their mind trying to use us, is more important than some
# memory usage double-ups, and just adding fields to a smart-object stops us *bogging down on that front*
# user-facing API usability is a lot more important than memory performance, especially for smallscale stuff like this.
# Licensing's "Entry and Registry-of-Entries" model seems to work quite well
# IF we're optimising for "give the user an opaque handle" operations, which we SHOULD ALWAYS be.

# ALWAYS copy-paste THEN DRY. Do NOT try to DRY in-flight!





# C3 private & public binary block structure saving, loading & validation

import copy, binascii, time, struct, os

from c3.constants import *
from c3.errors import StructureError, IntegrityError, CertExpired


class AttrDict(dict):
    def __getattr__(self, name):
        return self[name]
    def __deepcopy__(self, memo):
        return self.__class__({k: copy.deepcopy(v, memo) for k, v in self.items()})


# --- Private part/block save & load ---

# in:  private key bytes, and if they are unencrypted (bare) or not
# out: block bytes with header.

def make_priv_block(priv_bytes, bare=False):
    privd = AttrDict()
    privd["key_type"] = KT_ECDSA_PRIME256V1
    privd["priv_type"] = PRIVTYPE_BARE if bare else PRIVTYPE_PASS_PROTECT
    privd["priv_data"] = priv_bytes
    privd["crc32"] = binascii.crc32(privd.priv_data, 0) % (1 << 32)
    out_bytes = b3.schema_pack(PRIV_CRCWRAP_SCHEMA, privd)
    out_bytes_with_hdr = b3.encode_item_joined(PRIV_CRCWRAPPED, b3.DICT, out_bytes)
    return out_bytes_with_hdr


# in: block bytes from e.g. LoadFiles
# out: DICT with private key + metadata
# sanity & crc32 check the priv block, then shuck it and return the inner data.
# Caller must then decrypt the private key if needed.

def load_priv_block(block_bytes):
    _, index = expect_key_header([PRIV_CRCWRAPPED], b3.DICT, block_bytes, 0)
    privd = AttrDict(b3.schema_unpack(PRIV_CRCWRAP_SCHEMA, block_bytes[index:]))
    # --- Sanity checks ---
    schema_ensure_mandatory_fields(PRIV_CRCWRAP_SCHEMA, privd)
    if privd.priv_type not in KNOWN_PRIVTYPES:
        raise StructureError("Unknown privtype %d in priv block (wanted %r)" % (privd.priv_type, KNOWN_PRIVTYPES))
    if privd.key_type not in KNOWN_KEYTYPES:
        raise StructureError("Unknown keytype %d in priv block (wanted %r)" % (privd.key_type, KNOWN_KEYTYPES))
    # --- Integrity check ---
    data_crc = binascii.crc32(privd.priv_data, 0) % (1 << 32)
    if data_crc != privd.crc32:
        raise IntegrityError("Private key block failed data integrity check (crc32)")
    return privd


# --- Public part loader ---

# In: public_part bytes
# Out: data-and-sig list cert chain structure
# Note: the inverse of this function is part of sign

def load_pub_block(public_part):
    # The public part should have an initial header that indicates whether the first das is a payload or a cert
    ppkey, index = expect_key_header([PUB_CSR, PUB_PAYLOAD, PUB_CERTCHAIN], None, public_part, 0)
    public_part = public_part[index:]               # chop off the header

    # PUB_CSR is just a cert by itself
    if ppkey == PUB_CSR:
        cert = AttrDict(b3.schema_unpack(CERT_SCHEMA, public_part))
        return ppkey, cert
    # PUB_PAYLOAD and PUB_CERTCHAIN are chains.
    # Should be a list of DAS structures, so pythonize the list

    chain = list_of_schema_unpack(DATASIG_SCHEMA, [HDR_DAS], public_part)

    # unpack the certs & sigs in chain
    for i, das in enumerate(chain):
        # dont unpack cert dif this is the first das and ppkey is PAYLOAD
        if i > 0 or ppkey == PUB_CERTCHAIN:
            das["cert"] = AttrDict(b3.schema_unpack(CERT_SCHEMA, das.data_part))
            schema_ensure_mandatory_fields(CERT_SCHEMA, das.cert)

        das["sig"] = AttrDict(b3.schema_unpack(SIG_SCHEMA, das.sig_part))
        schema_ensure_mandatory_fields(SIG_SCHEMA, das.sig)

    return ppkey, chain


# --- Structure helper functions ---

# Expects a list items which are the same schema object. This should eventually be part of b3.

def list_of_schema_unpack(schema, want_keys, buf):
    end = len(buf)
    index = 0
    out = []
    while index < end:
        try:
            key, data_type, has_data, is_null, data_len, index = b3.item.decode_header(buf, index)
        except (IndexError, UnicodeDecodeError):
            raise StructureError("List item header structure is invalid")
        if key not in want_keys:
            raise StructureError \
                ("List item header key invalid - wanted %r got %r" % (want_keys, key))
        if data_type != b3.DICT:
            raise StructureError("List item header type invalid - wanted DICT got %r" % data_type)
        if not has_data or data_len == 0:
            raise StructureError("List item header invalid - no data")

        das_bytes = b3.decode_value(data_type, has_data, is_null, data_len, buf, index)

        if len(das_bytes) == 0:
            raise StructureError("List item data is missing")

        # Now unpack the actual dict too
        dx = b3.schema_unpack(schema, das_bytes)
        schema_ensure_mandatory_fields(schema, dx)
        out.append(AttrDict(dx))
        index += data_len
    return out


def schema_ensure_mandatory_fields(schema, dx):
    for field_def in schema:                    # by name
        # only check if mandatory bool flag is both present AND true.
        if len(field_def) > 3 and field_def[3] is True:
            field_name = field_def[1]
            if field_name not in dx:
                raise StructureError("Required schema field '%s' is missing" % field_name)
            if not dx[field_name]:
                raise StructureError("Mandatory field '%s' is %r" % (field_name, dx[field_name]))


# Index and Unicode are the only two unhandled exception types that b3's decode_header code produces when fuzzed.
# IndexError trying to decode a bad varint for ext_type, datalen or number key.
# Unicode for when b3 thinks there's a utf8 key but the utf8 is bad.

def expect_key_header(want_keys, want_type, buf, index):
    if not buf:
        raise StructureError("No data - buffer is empty or None")
    try:
        key, data_type, has_data, is_null, data_len, index = b3.decode_header(buf, index)
    except (IndexError, UnicodeDecodeError):
        raise StructureError("Header structure is invalid")  # from None
        # raise .. from None disables py3's chaining (cleaner unhandled prints) but isnt legal py2
    if key not in want_keys:
        raise StructureError \
            ("Incorrect key in header - wanted %r got %s" % (want_keys, repr(key)[:32]))
    if want_type is not None and want_type != data_type:   # note often we dont want type checking
        raise StructureError \
            ("Incorrect type in header - wanted %r got %s" % (want_type, repr(data_type)[:32]))
    if not has_data:
        raise StructureError("Invalid header - no has_data")
    if index == len(buf):
        raise StructureError("No data after header - buffer is empty")
    return key, index

# Used by the uniloader when someone passes a binary block directly.

def split_binary_pub_priv(block_in):
    want_keys = [PUB_CSR, PUB_PAYLOAD, PUB_CERTCHAIN, PRIV_CRCWRAPPED, DUALBLOCK]
    key, index = expect_key_header(want_keys, None, block_in, 0)
    if key == PRIV_CRCWRAPPED:
        return b"", block_in
    if key in (PUB_CSR, PUB_PAYLOAD, PUB_CERTCHAIN):
        return block_in, b""
    if key == DUALBLOCK:
        dual = b3.schema_unpack(DUALBLOCK_SCHEMA, block_in[index:])
        return dual["public"], dual["private"]

def combine_binary_pub_priv(pub_block, priv_block):
    dbx = dict(public=pub_block, private=priv_block)
    dual_bytes = b3.schema_pack(DUALBLOCK_SCHEMA, dbx)
    return b3.encode_item_joined(DUALBLOCK, b3.DICT, dual_bytes)


# --- Output/Results fetchers ---

# In: chain from load()
# Out: payload bytes

def get_payload(chain):
    first_das = chain[0]
    if "cert" not in first_das:         # first data_and_sig element's data_part is a payload
        return first_das.data_part
    return b""      # first data_and_sig element's data_part is a cert


# Delivering a cut-down Chain for meta:
# chain is list of:
#   data_part,  ->  cert (or payload)   ->  cert fields
#   sig_part,   ->  sig                 ->  signature & signing_cert_id

# In: chain from load()
# Out: trimmed-down metadata-only chain

def get_meta(chain):
    st = 0
    first_das = chain[0]
    if "cert" not in first_das:     # skip the first data_and_sig if it's a payload one
        st = 1
    chain2 = copy.deepcopy(chain[st:])
    for i in chain2:
        del i["data_part"]
        del i["sig_part"]
        del i["cert"]["public_key"]
        del i["sig"]["signature"]
    return chain2


# For error message readability
def ctnm(das):
    if not das:
        return ""
    if "cert" in das:
        return " (cert %r) " % das.cert.subject_name
    else:
        return " (payload) "

# --------------------------------------------------------------------------------------------------

# ULID generator that doesn't care about trying to be sub-millisecond monotonic.
def gen_ulid():
    stamp_bytes = struct.pack(">Q", int(time.time() * 1000.0))[2:]
    rand_bytes = os.urandom(10)
    return stamp_bytes + rand_bytes


# Note: a better approach here is to use micro or nanosecond timers.
# This guy has a number of good strategies:  https://github.com/RobThree/NUlid/tree/master/NUlid/Rng
# random vs monotonic is a tradeoff-spectrum, with cryptographically-secure RNG at one and and
# "just use DB autoincrement fields instead" at the other.


# ULID gen using super basic "keep trying random numbers" monotonic-increase tekneeq
# Dont want it sucking systems dry of urandom though.
# prev_rand = None
# prev_stamp = None
# def gen_ulid():
#     global prev_rand, prev_stamp
#     nn = 0
#     stamp = int(time.time() * 1000.0)
#     stamp_bytes = struct.pack(">Q", stamp)[2:]
#     rand_bytes = os.urandom(10)
#     if stamp == prev_stamp and prev_rand is not None:
#         while rand_bytes < prev_rand:
#             nn += 1
#             rand_bytes = os.urandom(10)
#     prev_rand = rand_bytes
#     prev_stamp = stamp
#     return stamp, nn, stamp_bytes + rand_bytes

# Courtesy of https://github.com/valohai/ulid2/blob/master/ulid2/__init__.py



from setuptools import setup, find_packages
import codecs
import os.path

here = os.path.abspath(os.path.dirname(__file__))

def read(rel_path):
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


setup(
    name="c3sign",
    version=get_version("c3/__init__.py"),
    packages=find_packages(),
    install_requires=["b3buf", "ecdsa"],
    description="Compact Crypto Certs (C3) is a mini-PKI signer/verifier with full chain functionality and compact binary and friendly text cert formats",
    long_description=open(os.path.join(here, "README.md"), "r").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/oddy/c3",
    author="Beau Butler",
    author_email="beau.butler@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
    ],
    include_package_data=True,
    data_files=[ ("", ["LICENSE.txt",],) ],
    # we want a universal wheel
    options={"bdist_wheel": {"universal": True}},

    entry_points={
        "console_scripts" : [
            "c3 = c3.__main__:main"
        ]
    }

)




from __future__ import unicode_literals     # for python2

import base64, traceback, random, os, datetime
from pprint import pprint

from c3.signverify import SignVerify


# ---- Truncate and glitch loops -----
# A 2-cert chain here.
root1_pub = """
2Q66AekZtgEJAGUJAAVyb290MRkBBXJvb3QxOQIBAQkDQDiZmemfIXf+5WpBaAJYOlgqitfC4txC
5kBdGEgaA0A+jblvdn5LdxdO0heNDDNBh9nYDZUJUwo+nOpyd/H05kWpBAQYCtAfqQUEFQvMHwkB
SwkAQEbrVdKzH6YpeQiKo9bKmHWUV0yIu6H4tsOfrGJhEnARutowDamrPlcBdT9kxqzUkauFU6Dd
+xyEFoYaLEOTrJ4JAQVyb290MQ==
"""
root1_block = base64.b64decode(root1_pub)
inter_pub = """
2Q7uAukZsAEJAGUJAAVpbnRlchkBBWludGVyOQIBAQkDQJqwMvo4wnp9Auz93zKrscBlbu8mihvz
RARudJyYDumHVutovutOUEaPiJuQjMpI/vOG+hxn6uhZv7/aTM+B+e+pBAQYCtAfqQUEFQvMHwkB
RQkAQMmN9FKivxybOhv+dkqU2qhhcr3DSSvKqdFF4jBBp3yjV4mT3JJ1mdJ2m8sJq9cIyeRAhzgF
kBJD7wBH/RnwrNQBAekZtgEJAGUJAAVyb290MRkBBXJvb3QxOQIBAQkDQDiZmemfIXf+5WpBaAJY
OlgqitfC4txC5kBdGEgaA0A+jblvdn5LdxdO0heNDDNBh9nYDZUJUwo+nOpyd/H05kWpBAQYCtAf
qQUEFQvMHwkBSwkAQEbrVdKzH6YpeQiKo9bKmHWUV0yIu6H4tsOfrGJhEnARutowDamrPlcBdT9k
xqzUkauFU6Dd+xyEFoYaLEOTrJ4JAQVyb290MQ=="""
inter_block = base64.b64decode(inter_pub)

# Testing what happens if the public_part buffer is incomplete
# (And finding out exactly where to truncate public_part for the short-chain test above)

# Run me from TOP top level with: "python -m c3.tests.test_c3"

# Because the public block as a whole has a size, we need to adjust that size also, otherwise
# B3 just errors out across the entire buffer (as it should).

def truncmain():
    c3m = SignVerify()
    #c3m.load_trusted_cert(block=root1_block)
    buf = inter_block[:]

    for i in range(len(buf)+1, 1, -1):
        buf2 = buf[:i]
        try:
            xx = c3m.load(block=buf2)
        except Exception as e:
            print("%4i    load  %r" % (i,  e))
            #print(traceback.format_exc())
            continue
        try:
            c3m.verify(xx)
        except Exception as e:
            print("%4i  verify  %r" % (i, e))
            continue
        print("%4i   - SUCCESS -" % (i,))


# glitch a byte anywhere? in the chain to trigger signature fails.

def glitchmain():
    c3m = SignVerify()
    c3m.load_trusted_cert(block=root1_block)
    buf = inter_block[:]

    for i in range(len(buf)):
        buf2 = buf[:i] + b"\x00" + buf[i+1:]
        try:
            xx = c3m.load(block=buf2)
        except Exception as e:
            print("%4i    load   %20s" % (i,e))
            if "index out of" in str(e):
                print()
                print(traceback.format_exc())
                print()
            continue
        try:
            c3m.verify(xx)
        except Exception as e:
            print("%4i  verify   %r" % (i,e))
            continue
        print("%4i   - SUCCESS -" % (i,))

def smallrandfuzz():
    c3m = SignVerify()
    z = {}
    i = 0
    while True:
        i += 1
        buf = random.randbytes(40)
        #buf = b"\xdd\x37\x40\xed\x4d\x30\x44" + random.randbytes(60)
        try:
            xx = c3m.load_pub_block(buf)
            out = "omg SUCCESS omg"
        except Exception as e:
            out = str(e)
        z[out] = z.get(out,0) + 1

        if i % 100000 == 0:
            print()
            pprint(z)
            return



if __name__ == '__main__':
    truncmain()
    #glitchmain()
    #smallrandfuzz()
    #interactive_password_test()



# ---- Basic fuzzing of the initial header check ----
#
# def FuzzEKH():
#     for i in range(0,255):
#         buf = six.int2byte(i) #+ b"\x0f\x55\x55"
#         try:
#             ppkey, index = expect_key_header([PUB_PAYLOAD, PUB_CERTCHAIN], b3.LIST, buf, 0)
#             print("%4i %02x - SUCCESS - key = %r" % (i,i, ppkey))
#         except Exception as e:
#             print("%4i %02x -  %s" % (i,i, e))
#             #print(traceback.format_exc())
#
# def FuzzEKH2():
#     i = 0
#     z = {}
#     while True:
#         i += 1
#         buf = random.randbytes(20)
#         try:
#             ppkey, index = expect_key_header([PUB_PAYLOAD, PUB_CERTCHAIN], b3.LIST, buf, 0)
#             out = "SUCCESS - key = %r" % ppkey
#         except Exception as e:
#             out = "%r" % e
#
#         #print(out)
#         z[out] = z.get(out,0) + 1
#
#         if i % 100000 == 0:
#             print()
#             print(len(z))
#             pprint(z)
